# Generated by Django 5.1 on 2024-09-25 04:04

from django.db import migrations
from django.db import models
import django.db.models.deletion


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
    ]

    operations = [
        migrations.CreateModel(
            name="AsyncTempFile",
            fields=[
                (
                    "id",
                    models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID"),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("original_filename", models.TextField()),
                ("key", models.CharField(max_length=500)),
                ("field_name", models.CharField(max_length=100)),
                ("moved_to_location", models.TextField(blank=True)),
                ("error", models.TextField(null=True)),
                (
                    "content_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="contenttypes.contenttype"
                    ),
                ),
            ],
            options={
                "db_table": "alliance_platform_storage_async_temp_file",
            },
        ),
    ]
